/*
 * Ticker.h
 *
 *  Created on: Jan 6, 2016
 *      Author: eric
 */

#ifndef INCLUDE_FASTEC_INTERNAL_SYS_TICKER_H_
#define INCLUDE_FASTEC_INTERNAL_SYS_TICKER_H_

#include <fastec/fas_trace.h>

#include <fastec/interval_timer.h>
#include <fastec/port.h>

#ifdef _WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <condition_variable>
#include <mutex>
#include <thread>

namespace Fastec
{

/**
 * @brief Ticker class
 *
 * The ticker arranges to call its tick() method at regular (periodic) intervals.
 * The interval is given as a parameter to the constructor. The Ticker establishes
 * its own independent thread of execution, and executes the tick()
 * method on that thread. Ticker is a virtual class. A derived class must implement
 * the virtual tick() method.
 *
 * A ticker is initially stopped, and must be started by a call to the start()
 * before tick() executions will begin.
 *
 * A Ticker maintains statistics on its execution (see period and sleeptime
 * data members).
 */
class Ticker : public Trace
{
public:
    static void threadEntry(Ticker* t)
    {
        t->hold();
        t->runTicker();
    }

    /** @name Trace control @{ */
    static const uint32_t TRACE_SEL_TIMER = 0x00000008; /**< @brief Trace timer operations */
    /***}*/

    /**
     * @brief Constructor for Ticker
     *
     * @param t Ticker's period, in microseconds
     */
    Ticker(unsigned t);
    virtual ~Ticker();
    void runTicker( void );

    /**
     * @brief Change the time period
     *
     * @param t  New time period, in microseconds
     */
    virtual void setTime(unsigned t);

    /**
     * @brief Start the Ticker (begin calling tick() periodically).
     */
    virtual void start();

    /**
     * @brief Stop the Ticker (no more calls to tick until after another start()).
     */
    virtual void stop();

    /**
     * @brief Pure virtual tick() method
     *
     * Concrete sub-classes must provide an implementation of the tick() method.
     * The Ticker will execute the tick() method once every 'interval' microseconds.
     */
    virtual void tick() = 0;

    /**
     * @brief Track min/mean/max intervals between tick() executions
     */
    IntervalTimer period;

    /**
     * @brief Track min/mean/max sleep times
     *
     * This timer tracks the times that the Ticker thread spends blocked, awaiting
     * the next call to tick(). Time starts after one execution of tick() and ends
     * just before the next call to tick(). This can be used to judge how much of
     * the Ticker's interval is actually consumed by the tick() method's execution
     * time.
     */
    IntervalTimer sleepTime;

protected:
    unsigned interval;

private:
    void initTimer();
    void unInitTimer();
    void cancelTimer();
    void endTimer();

    std::mutex *mutex;
    bool running;
    int tickerFd;
    std::thread *tickerThread;

    //! @name Thread startup synchronization
    //!
    //! Mechanism to make sure thread does not brgin
    //!
    //! @{

    //! @brief Hold at thread startup until released
    //!
    //! This is intended for use by a thread's startup function. Thread
    //! execution will block here until released by a call to the
    //! @ref release "release()" method.
    //!
    void hold( void );

    //! @brief Release thread for execution
    //!
    //! This is intended for use by sub-class constructors that have
    //! created a thread using the _hold_ = true parameter. This will
    //! allow the new thread, blocked in @ref hold "hold()", to continue
    //! execution. Sub-class constructors can use this to prevent
    //! threads' _run()_ methods from accessing derived class members
    //! before they are initialized.
    //!
    void release( void );

    //! @brief A mutex to go with the condition variable
    std::mutex	*startup_mutex;

    //! @brief A condition variable to go with the mutex
    std::condition_variable *startup_condition;

    //! @brief State variable for the thread to test
    bool startup_ok;
    //! @}

#ifdef __linux__
    //! @name Ticker thread signaling mechanism for Linux
    //! @{
    int signal_fd[2];
    int max_fd_plus_one;
    //! @}
#endif // __linux__

#ifdef _WIN32
private:
    static const int NOP = 0;
    static const int START = 1;
    static const int STOP = 2;
    int cmd;

    HANDLE timerHandle;
    HANDLE timerThreadHandle;

    bool timerThreadRunning;
    std::mutex *tickerThreadLock;
    std::condition_variable *tickerThreadCond;
#endif // _WIN32
};

} /* namespace Fastec */

#endif /* INCLUDE_FASTEC_INTERNAL_SYS_TICKER_H_ */
